// FormPostParser.cpp : Implementation of CFormPostParser
#include "stdafx.h"
#include "FormPostBatchParser.h"
#include "FormPostParser.h"
#include <algorithm>

/////////////////////////////////////////////////////////////////////////////
// CFormPostParser
//

/////////////////////////////////////////////////////////////////////////////
// Stream helpers
//   loadStream() -- loads data from a stream in dynamically allocated buffer
//   peekStream() -- just like loadStream() but doesn't change current read position
//
//  Both return dynamically allocated buffers.  Caller frees using operator delete[].
//  Both accept nRequestedSize which can be (size_t)-1 to get entire stream.
//  Both have an optional 3rd parameter for the number of bytes actual read.
//
static BYTE * loadStream ( IStream * pstm, size_t nRequestedSize, size_t & nActualSize )
{
    nActualSize = 0; // default value for output parameter
    if (! pstm) return 0;

    // determine the entire stream size, if no size is provided
    if (nRequestedSize == (size_t)-1)
    {
        STATSTG stat; 
        memset(&stat, 0, sizeof stat);
        if (FAILED(pstm->Stat(&stat, STATFLAG_NONAME))) return 0;
        nRequestedSize = (size_t)stat.cbSize.QuadPart;
    }

    // allocate a buffer and clear to zero
    BYTE * buffer = new unsigned char[2 + nRequestedSize];
    memset(buffer, 0, 2+nRequestedSize); // extra L'\0' in case being null-terminated is useful

    // read in the buffer and record how many bytes actually read
    ULONG nBytesRead = 0;
    HRESULT hr = pstm->Read(buffer, (ULONG)nRequestedSize, &nBytesRead);
    if (FAILED(hr))
    {
        delete[] buffer;
        buffer = 0;
    }

    // set the number of bytes actually read
    if (buffer)
        nActualSize = (size_t) nBytesRead;

    return buffer;
}

static BYTE * loadStream ( IStream * pstm, size_t nRequestedSize = (size_t) -1 /* defaults to entire stream size */ )
{
    size_t nUnused;
    return loadStream(pstm, nRequestedSize, nUnused);
}

static BYTE * peekStream ( IStream * pstm, size_t nPeek, size_t & nPeeked )
{
    // load the requested number of bytes from the stream
    BYTE * buffer = loadStream(pstm, nPeek, nPeeked);

    // restore read position so the read is really a peek
    if (buffer && (nPeeked > 0))
    {
        LARGE_INTEGER nOffset; nOffset.QuadPart = - (LONGLONG)nPeeked;
        ATLASSERT(SUCCEEDED(pstm->Seek(nOffset, STREAM_SEEK_CUR, 0)));
    }

    return buffer;
}

static BYTE * peekStream ( IStream * pstm, size_t nPeek = (size_t) -1 /* defaults to entire stream size */ )
{
    size_t nUnused;
    return peekStream(pstm, nPeek, nUnused);
}


/////////////////////////////////////////////////////////////////////////////
// ATLTRACE helpers
//   _TRACER(x) -- returns a non-null string value for various types x.
//                 The return type is either BSTR or CComBSTR.
//   _AtlTraceBigBufer() -- a replacement for ATL's ATLTRACE mechanism
//                 which has a limit of 512 chars and fires an assert
//                 if the message is greater than 512.  Firing an 
//                 assert message box inside a component loaded into a 
//                 service is not good.
#ifdef _DEBUG
#undef ATLTRACE
#define ATLTRACE _AtlTraceBigBuffer

void _AtlTraceBigBuffer ( LPCTSTR pszFormat, ... )
{
	const size_t nSize = 8192;
	TCHAR * sz = new TCHAR[nSize];
	memset(sz, 0, nSize);
	va_list args;
	va_start(args, pszFormat);
	_vsntprintf(sz, nSize-1, pszFormat, args);
    va_end(args);
	OutputDebugString(sz); //would call ATLTRACE but szReg is > 512 bytes
	delete[] sz;
}
static BSTR _TRACER ( BSTR s )
{
    return s ? s : L"";
}

static CComBSTR _TRACER ( IStream * pstm, size_t nPrefixLen = 50 ) // nPrefixLen = how many bytes of the stream prefix should be returned
{
    if (! pstm) return L"<null>";
    wchar_t wsz[4096];
    wcscpy(wsz, L"{size:");
    STATSTG stat;
    memset(&stat, 0, sizeof stat);
    if (SUCCEEDED(pstm->Stat(&stat, STATFLAG_NONAME)))
        swprintf(wsz+wcslen(wsz), L"%I64u", stat.cbSize.QuadPart);
    else
        wcscat(wsz, L"<unavailable>");
    if (nPrefixLen)
    {
        wcscat(wsz, L" data:");
        // peek at data; get 1 more byte than nPrefixLen so can test if there is more data
        size_t nBytesRead;
        char * pszData = (char *)peekStream(pstm, nPrefixLen+1, nBytesRead);
        bool bMoreData = (nBytesRead > nPrefixLen);
        if (bMoreData) -- nBytesRead; // don't use the extra byte read just for testing purposes
        if (! pszData)
            wcscat(wsz, L"<unavailable>");
        else
        {
            wcscat(wsz, L"\"");
            for ( int i = 0;  i < (int)nBytesRead;  ++i )
                if (isprint((unsigned char)pszData[i]))
                    swprintf(wsz+wcslen(wsz), L"%C", pszData[i]);
                else
                    switch (pszData[i])
                    {
                        case '\0': wcscat(wsz, L"\\0"); break;
                        case '\r': wcscat(wsz, L"\\r"); break;
                        case '\n': wcscat(wsz, L"\\n"); break;
                        case '\t': wcscat(wsz, L"\\t"); break;
                        case '\"': wcscat(wsz, L"\\\""); break;
                        default:   swprintf(wsz+wcslen(wsz), L"\\x%02X", (int)(unsigned char)pszData[i]); break;
                    }
            if (bMoreData) wcscat(wsz, L"...");
            wcscat(wsz, L"\"");
        }
        delete[] (BYTE*)pszData;
    }
    wcscat(wsz, L"}");
    return wsz;
}

static CComBSTR _TRACER ( IDictionary * pdict )
{
    if (! pdict) return L"<null>";
    // get the number of items in the dictionary
    wchar_t wsz[8192];
    wcscpy(wsz, L"{count:");
    long nCount = 0;
    if (FAILED(pdict->get_Count(&nCount)))
        wcscat(wsz, L"<unavailable>");
    else
    {
        swprintf(wsz+wcslen(wsz), L"%ld", nCount);
        if (nCount > 0) 
        {
            wcscat(wsz, L" items:");
            CComPtr<IUnknown> spunkEnum;
            if (FAILED(pdict->get__NewEnum(&spunkEnum))) spunkEnum = 0;
            CComQIPtr<IEnumVARIANT> spEnum(spunkEnum);
            if (! spEnum)
                wcscat(wsz, L"<unavailable>");
            else
                for ( ;; )
                {
                    CComVariant vaKey, vaValue;
                    if (S_OK != spEnum->Next(1, &vaKey, 0)) break;
                    if (FAILED(vaKey.ChangeType(VT_BSTR))) continue;
                    swprintf(wsz+wcslen(wsz), L"[%ls]=", vaKey.bstrVal ? vaKey.bstrVal : L"");
                    if (SUCCEEDED(pdict->get_Value(vaKey.bstrVal, &vaValue)) && SUCCEEDED(vaValue.ChangeType(VT_BSTR)))
                        swprintf(wsz+wcslen(wsz), L"\"%ls\" ", vaValue.bstrVal ? vaValue.bstrVal : L"");
                    else
                        wcscat(wsz, L"<unavailable>");
                }
        }
    }
    wcscat(wsz, L"}");
    return wsz;
}
#endif



/////////////////////////////////////////////////////////////////////////////
// IBizTalkParserComponent implementation
//

// IBizTalkParserComponent::ProbeInterchangeFormat()
// -------------------------------------------------
//  identify whether the given data stream matches the format handled by this parser.
//
//  Called before any other member of this interface.  No other interface member is called
//  if this function indicates the data stream is not handled by the parser.
//
//  on input,
//      pstmData = stream containing the data for the parser.  Read-only.  Only members
//                 Read(), Seek(), and Stat() are defined.  The current position in
//                 the stream is not necessarily zero.
//      bFromFile = TRUE, indicates pstmData was loaded from a file and no assumption can be made about format
//                  FALSE, indicates pstmData was initialized from a BSTR which indicates UNICODE format
//      bstrEnvName = name of the associated envelope or null if there is none.  Is the value
//                    passed into IInterchange (see below)
//      pstmReceiptData = stream containing the receipt document
//
//      The first three parameters are derived from the parameters to IInterchange::Submit() 
//      and IInterchange::SubmitSync().  Those two functions are the only way to get an
//      an interchange submitted to BizTalk.  IInterchange accepts the data for the interchange
//      as a BSTR or a file reference.  pstmData is an IStream on either data source and 
//      bFromFile indicates which kind of data source (file or string).
//
//  on output,
//      pbstrFormat = document format.  Valid for IBizTalkDocument::Type property which must have
//                    one of the values "x12", "edifact", "custom xml", "custom", "flatfile", or "reliable".
//                    This value is set to null if S_FALSE is returned.
//
//  return value
//      S_OK, if the format is handled by this parser.  pbstrFormat is set.
//      S_FALSE, if the format is not handled by this parser.  pbstrFormat is set to null.
//      E_POINTER if input parameter pointers are null
// 
//  implementation note:
//      Per the BizTalk documention, pstmData is addrefed and stored if S_OK is returned.
//
STDMETHODIMP CFormPostParser::ProbeInterchangeFormat (IStream * pstmData, BOOL bFromFile, BSTR bstrEnvName, IStream * pstmReceiptData, BSTR * pbstrFormat)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::ProbeInterchangeFormat([in] pstmData=%ls, [in] bFromFile=%ld, [in] bstrEnvName=\"%ls\", [in] pstmReceiptData=%ls)\n"),
             (BSTR)_TRACER(pstmData), bFromFile, (BSTR)_TRACER(bstrEnvName), (BSTR)_TRACER(pstmReceiptData));
    if (!pstmData || !pbstrFormat) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::ProbeInterchangeFormat(), return E_POINTER\n");
        return E_POINTER;
    }

    // Decide whether can handle this format and set local variable bCanHandleFormat accordingly.
    //  The stock implementation here tests if the stream begins with a schema identifier.
    //  Care is taken to not change the current stream position.
    const wchar_t SCHEMAID[] = L"\xfeff" L"schema=biztalk-unleashed-custom-parser-ch18";
    size_t nSchemaLen = sizeof SCHEMAID - 2, nPeeked = 0;
    wchar_t * buffer = (wchar_t*)::peekStream(pstmData, nSchemaLen, nPeeked);
    bool bCanHandleFormat = buffer && (nPeeked == nSchemaLen)
                            && ((0 == ::memcmp(buffer, SCHEMAID, nSchemaLen)) ||
								(0 == ::memcmp(buffer, SCHEMAID+1, nSchemaLen-2)));
	delete[] (BYTE*)buffer;

    // return whether this parser recognizes the format
    // if true is returned, then a copy of the data stream is retained for subsequent use
    if (! bCanHandleFormat)
    {
        *pbstrFormat = 0;
        ATLTRACE("\nExit IBizTalkParserComponent::ProbeInterchangeFormat([out] pbstrFormat=%ls), return S_FALSE\n", _TRACER(*pbstrFormat));
        return S_FALSE;
    }

    *pbstrFormat = CComBSTR(L"custom").Detach();

    // cache the stream for the interchange for later use
    m_spstmData = pstmData;
    ATLTRACE("\nExit IBizTalkParserComponent::ProbeInterchangeFormat([out] pbstrFormat=\"%ls\"), return S_OK\n", _TRACER(*pbstrFormat));
    return S_OK;
// ---------------------------------------------------------------------------
// Changes from ATL Object Wizard
// 1. Initalized local variable SCHEMAID.
// 2. Deleted pragma.
//
// Sample trace on entry to this function:
//     Enter IBizTalkParserComponent::ProbeInterchangeFormat(
//           [in] pstmData={size:369 data:"schema=biztalk-unleashed-custom-parser-ch18\r\nbroke"}, 
//           [in] bFromFile=1, [in] bstrEnvName="", [in] pstmReceiptData={size:0 data:""})
}

// IBizTalkParserComponent::GetInterchangeDetails()
// ------------------------------------------------
//  receive information about the interchange such as the envelope name, if any,
//  and identifiers for the source and destination organizations.
//
//  on input,
//      pdict = dictionary of <field,value> pairs.  See the BizTalk documentation 
//              for the fields stored in the dictionary.
// 
STDMETHODIMP CFormPostParser::GetInterchangeDetails (IDictionary * pdict)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetInterchangeDetails([in] pdict=%ls)\n"), (BSTR)_TRACER(pdict));
    if (! pdict) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetInterchangeDetails(), return E_POINTER\n");
        return E_POINTER;
    }
    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetInterchangeDetails(), return S_OK\n"));
    return S_OK;
// ---------------------------------------------------------------------------
// Changes from ATL Object Wizard
// 1. 
//
// Sample trace on entry to this function:
//     Enter IBizTalkParserComponent::GetInterchangeDetails(
//           [in] pdict={count:5 items:[Tracking_ID]="{316D42A9-769F-404C-9F52-8D6BFE7049C8}" 
//                                     [src_filepath]="e:\temp\CustomParserBatchReceive" 
//                                     [Document_Name]="definitionInboundBrokerBatch_ch18" 
//                                     [submission_id]="{C87F22B6-703C-414A-A762-4C95F7FC05A3}" 
//                                     [src_filename]="batch2.txt"})
}

// IBizTalkParserComponent::GroupsExist()
// --------------------------------------
//  flag whether the interchange contains groups.  Called before any 
//  parsing takes place.
//
//  An interchange can be a single document, a sequence of documents, or 
//  a sequence of groups.  A group itself contains a sequence of documents.
//
//  Called before any parsing takes place, immediately after ProbeInterchangeFormat().
//
//  on output,
//      pbGroupsExist = whether there are groups
// 
STDMETHODIMP CFormPostParser::GroupsExist (BOOL * pbGroupsExist)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GroupsExist()\n"));
    if (! pbGroupsExist) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GroupsExist(), return E_POINTER\n");
        return E_POINTER;
    }
    *pbGroupsExist = TRUE;
    ATLTRACE(_T("\nExit IBizTalkParserComponent::GroupsExist([out] pbGroupsExist=%ld), return S_OK\n"), *pbGroupsExist);
    return S_OK;
// ---------------------------------------------------------------------------
// Changes from ATL Object Wizard
// 1. Set *pbGroupsExist = TRUE
// 2. Deleted pragma
}

// IBizTalkParserComponent::GetGroupDetails()
// ------------------------------------------
//  specifies details for a single group to be used in the BizTalk Tracking Database.
//  Called before each group gets processed.
//
//  on input,
//      pdict = dictionary of <field,value> pairs.  See the article
//              "Supporting the Tracking Database with Parser and Serializer Components"
//              in the BizTalk documentation for the fields stored in the dictionary.
// 
struct lesspwsz // STL comparitor using wcscmp
{
    bool operator() (const wchar_t * x, const wchar_t * y) const { return ::wcscmp(x, y) < 0; }
};
STDMETHODIMP CFormPostParser::GetGroupDetails (IDictionary * pdict)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetGroupDetails([in] pdict=%ls)\n"), (BSTR)_TRACER(pdict));
    if (! pdict) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetGroupDetails(), return E_POINTER\n");
        return E_POINTER;
    }

    // read the entire stream and parse it, on first access
    //      This parsing is necessary so can deal with interleaved groups.
    //      Example data:
    //          schema=biztalk-unleashed-custom-parser-ch18
    //          broker=Delta%20Partners&stock=WXYZ&quantity=1000&min-price=25.00
    //          broker=Beta%20Brokerage&stock=JKL&quantity=1500&min-price=35.00
    //          broker=Beta%20Brokerage&stock=WXYZ&quantity=2500&min-price=100.00
    //          broker=Alpha%20Financials&stock=DEF&quantity=2500&min-price=20.00
    //          broker=Delta%20Partners&stock=STUV&quantity=50&min-price=35.00
    //      Grouping is done on the 'broker' field.
    //
    // To parse:  Documents are newline-delimted.  Fields in a document are &-delimited.
    if (! m_pwchStream)
    {
        // load the entire stream of newline-delimited documents
        m_pwchStream = (wchar_t*)loadStream(m_spstmData);
        for ( wchar_t * s = ::wcstok(m_pwchStream, L"\r\n");  s;  s = ::wcstok(0, L"\r\n") ) 
        {   // s has the form "broker=xxx&field1=val1&field2=val2&..."
            wchar_t * x = ::wcschr(s, L'&'); // find first '&' to delimit group id 'broker=xxx'
            if (! x) continue; // every document line will have the '&'; the schema line will not;
            *x = L'\0'; // embed null between group id and field list
            m_docs.push_back(s); // push "broker=xxx\0filed1=val1&field2=val2&..."
        }
        std::sort(m_docs.begin(), m_docs.end(), lesspwsz()); // sort doc list by group
        m_docpos = m_docs.begin();
    }

    // given:  *m_docpos = 1st document in a group
    //         It has the form:  broker=xxx\0field1=val1&field2=val2&...&fieldn=valn\0

    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetGroupDetails([out] pdict=%ls), return S_OK\n"), (BSTR)_TRACER(pdict));
    return S_OK;
// ---------------------------------------------------------------------------
// Changes from ATL Object Wizard
// 1. Added caching of entire stream
}



// IBizTalkParserComponent::GetGroupSize()
// ---------------------------------------
//  specify group-level information for the BizTalk Tracking database.  Called after
//  the last document in a group has been parsed.
//
//  on output,
//      pnGroupSize = total size of the group
//      pbLastGroup = whether this is the last group.  FALSE means another group will be parsed.
//
static size_t docSize ( const wchar_t * d ) // return size in bytes
{
    // *d the form broker=xxx\0field1=val1&field2=val2&...&fieldn=valn\0
    //      There is an embedded null so wcslen() must be called twice.
    if (! d) return 0;
    size_t n = wcslen(d) + 1; // tally broker field plus embedded null following it
    n += wcslen(d+n);         // tally remaining fields after embedded null
	return (sizeof wchar_t) * n;
}
STDMETHODIMP CFormPostParser::GetGroupSize (long * pnGroupSize, BOOL * pbLastGroup)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetGroupSize()\n"));
    if (!pnGroupSize || !pbLastGroup) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetGroupSize(), return E_POINTER\n");
        return E_POINTER;
    }

    // at this point, m_docpos is advanced one position beyond the last document parsed

    // on the last group if have reached the end of the document list
    *pbLastGroup = (m_docpos == m_docs.end()) ? TRUE : FALSE;

    // walk backwards through the document list to tally the size of each
    //  document in the group just completed
    *pnGroupSize = 0;
	for ( int i0 = m_docpos - m_docs.begin() - 1, i = i0;
          (i >= 0) && (0 == ::wcscmp(m_docs[i], m_docs[i0]));  -- i )
        *pnGroupSize += ::docSize(m_docs[i]);

    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetGroupSize([out] pnGroupSize=%ld, [out] pbLastGroup=%ld), return S_OK\n"), *pnGroupSize, *pbLastGroup);
    return S_OK;
// ---------------------------------------------------------------------------
// Changes from ATL Object Wizard
// 1. Added setting of output parameters 
// 2. Deleted pragma.
}

// IBizTalkParserComponent::GetNextDocument()
// ------------------------------------------
//  parse the next document from the data stream.  Recall that the data stream was stored
//  as a class member in the call to ProbeInterchangeFormat().
//
//  on input,
//      pdict = dictionary into which to place results
//      bstrDocName = name of the document specification
//
//  on output,
//      pdict is updated.  Field "WORKING_DATA " contains the XML for the parsed document.
//                         Sets fields for routing if the document is self-routing.  These fields
//                         are "SrcQual", "SrcID", "DestQual", and "DestID".  
//                         Recall:  Key names are not case sensitive.
//      pbDocIsValid = whether the document instance is valid.  If false is returned, the read position
//                     on the stream must be advanced to the start of the next document.  If this error
//                     recovery is not possible, then E_FAIL must be returned to abort the interchange.
//      pbLastDocument = whether this doucment is the last in the current group or the entire interchange
//      pnReceiptGenerated = receipt indicator.
//      pbDocIsReceipt = whether the document is a receipt
//      pbstrCorrelationCompProgID = prog id for the correlation of documents
//
//  return value
//      S_OK -- document was parsed or if an error was detected, the stream position was advanced
//              to the start of the next document
//      E_FAIL -- if parsing the document failed and the interchange should be aborted
//
static int parseFormPostField ( const wchar_t * s, std::wstring & field, std::wstring & value )
{   // syntax:  field=value and is delimited on right by null or '&'
    //          field and value can contain embedded %xx escape sequences
    field.erase(); value.erase();
    const wchar_t * s0 = s;
    for ( std::wstring * t = & field;  *s && (*s != L'&');  ++s )
        if (*s == L'=') // advance to value portion of syntax
            t = & value;
        else if ((*s == L'%') && iswxdigit(s[1]) && iswxdigit(s[2])) // found hex character encoding such as '%20'
        {
            int asciicode;
            if ((1 == swscanf(s+1, L"%2x", &asciicode)) && asciicode)
                *t += btowc(asciicode);
            s += 2; // skip digits
        }
        else
            *t += *s;
    return s - s0; // number of characters parsed
}
static std::wstring xmlEscape ( const std::wstring s )
{
    std::wstring x;
    for ( std::wstring::const_iterator it = s.begin();  it != s.end();  it ++ )
    {
        switch (*it)
        {
            case L'&': x += L"&amp;"; break;
            case L'<': x += L"&lt;"; break;
            case L'>': x += L"&rt;"; break;
            case L'\"': x += L"&quot;"; break;
            case L'\'': x += L"&apos;"; break;
            default: x += *it; break;
        }
    }
    return x;
}
STDMETHODIMP CFormPostParser::GetNextDocument (IDictionary * pdict, BSTR bstrDocName, BOOL * pbDocIsValid, BOOL * pbLastDocument, enum GeneratedReceiptLevel * pnReceiptGenerated, BOOL * pbDocIsReceipt, BSTR * pbstrCorrelationCompProgID)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetNextDocument([in] pdict=%ls, [in] bstrDocName=\"%ls\")\n"), (BSTR)_TRACER(pdict), (BSTR)_TRACER(bstrDocName));
    if (!pdict || !pbDocIsValid || !pbLastDocument || !pnReceiptGenerated || !pbDocIsReceipt || !pbstrCorrelationCompProgID)
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetNextDocument(), return E_POINTER\n");
        return E_POINTER;
    }

    // set default output parameters
    *pbDocIsValid = FALSE;
    *pbLastDocument = FALSE;
    *pnReceiptGenerated = NoReceiptGenerated;
    *pbDocIsReceipt = FALSE;
    *pbstrCorrelationCompProgID = 0;

    // parse current document which is *m_docpos
    //     doc format:  broker=xxx\0field1=val1&field2=val2&...&fieldn=valn\0
    if (m_docpos == m_docs.end()) return S_FALSE; // unexpected end --> return invalid document (e.g., empty document sequence)
    wchar_t wsz[4096];
    ::wcscpy(wsz, L"<broker-item>"); // open xml document
    int n;
    for ( wchar_t * doc = *m_docpos;  *doc;  doc += n )
    {
        std::wstring field, value;
        n = ::parseFormPostField(doc, field, value);
        ::swprintf(wsz+::wcslen(wsz), L"<%s>%s</%s>", field.c_str(), ::xmlEscape(value).c_str(), field.c_str());
        if ((doc[n] == L'&') || (doc == *m_docpos)) n++; // skip '&' or embedded null after 1st field
    }
    ::wcscat(wsz, L"</broker-item>"); // close xml document

    // record the document
    *pbDocIsValid = TRUE;
    HRESULT hr = pdict->put_Value(L"working_data", CComVariant(wsz));
    ATLASSERT(SUCCEEDED(hr));

    // test if are on the last document in its group
    //      true if reached end of doc list or next doc has different "broker=xxx" field
    ++ m_docpos;
    if ((m_docpos == m_docs.end()) || (0 != ::wcscmp(*m_docpos, *(m_docpos-1))))
        *pbLastDocument = TRUE;

    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetNextDocument([out] pdict=%ls, [out] pbDocIsValid=%ld, [out] pbLastDocument=%ld, [out] pnReceiptGenerated=%ld, [out] pbDocIsReceipt=%ld, [out] pbstrCorrelationCompProgID=\"%ls\"), return S_OK\n"), 
                (BSTR)_TRACER(pdict), *pbDocIsValid, *pbLastDocument, (long)*pnReceiptGenerated, *pbDocIsReceipt, (BSTR)_TRACER(*pbstrCorrelationCompProgID));
    return S_OK;
// ---------------------------------------------------------------------------
// Changes from ATL Object Wizard
// 1. Added the code between the ATLTRACE statements.
// 2. Added 2 static functions parseFormPostField() xmlEscape().
// 3. Deleted pragma.
//
// Sample trace on entry to this function:
//     Enter IBizTalkParserComponent::GetNextDocument(
//           [in] pdict={count:6 items:[Tracking_ID]="{316D42A9-769F-404C-9F52-8D6BFE7049C8}" 
//                                     [src_filepath]="e:\temp\CustomParserBatchReceive" 
//                                     [Document_Name]="definitionInboundBrokerBatch_ch18" 
//                                     [syntax]="custom" 
//                                     [submission_id]="{C87F22B6-703C-414A-A762-4C95F7FC05A3}" 
//                                     [src_filename]="batch2.txt"}, 
//           [in] bstrDocName="definitionInboundBrokerBatch_ch18")
}

// IBizTalkParserComponent::GetNativeDocumentOffsets ()
// ----------------------------------------------------
//  specify the size and position of the unparsed data for the last document parsed
//  The requirement here is that the unparsed data be a contiguous block in the stream.
//
//  In contrast, the groups do not have to be contiguous and documents within a group
//  do not have to appear in any particular order.  The parsing has the opportunity to
//  reorganize the input stream into groups and documents within each group.  However, the
//  data for each each document must be contiguous.
//
//  on output,
//      pbSizeFromXMLDoc = TRUE,  if BizTalk should infer the size and position.  In this case,
//                                the other 2 output parameters are not used.
//                         FALSE, if the size and position are specifed in the other 2 output parameters.
//
//                         Inference is done based on the XML returned from GetNextDocument().
//      pnStartOffset = the offset within the stream for the start of the document's data in the unparsed form
//                      set only if pbSizeFromXMLDoc is set to FALSE
//      pnDocLength = the size within the stream for the document's data
//                    set only if pbSizeFromXMLDoc is set to FALSE
//
STDMETHODIMP CFormPostParser::GetNativeDocumentOffsets (BOOL * pbSizeFromXMLDoc, LARGE_INTEGER * pnStartOffset, long * pnDocLength)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetNativeDocumentOffsets()\n"));
    if (!pbSizeFromXMLDoc || !pnStartOffset || !pnDocLength)
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetNativeDocumentOffsets(), return E_POINTER\n");
        return E_POINTER;
    }

    // calculate size and offset of last document
    *pbSizeFromXMLDoc = FALSE; // false means we calculate size and offset here
    const wchar_t * pwchLastDoc = *(m_docpos - 1);
    pnStartOffset->QuadPart = (char*)pwchLastDoc - (char*)m_pwchStream; // offset from start of the stream
    *pnDocLength = ::docSize(pwchLastDoc);

    #if _DEBUG
        ATLTRACE(_T("\nExit IBizTalkParserComponent::GetNativeDocumentOffsets([out] pbSizeFromXMLDoc=%ld"), *pbSizeFromXMLDoc);
        if (! *pbSizeFromXMLDoc)
            ATLTRACE(_T(", [out] pnStartOffset=%I64d, [out] pnDocLength=%ld"), pnStartOffset->QuadPart, *pnDocLength);
        ATLTRACE(_T(")\n"));
    #endif
    return S_OK;
// ---------------------------------------------------------------------------
// Changes from ATL Object Wizard
// 1. Added the "calculate" block 
// 2. Deleted pragma.
}
