///////////////////////////////////////////////////////////////////////////////
// ifunctoidimpl.h -- a map-based implementation of IFunctoid.  
//
//  Provided by Bill Martschenko as part of the book BizTalk Unleashed.
//  All rights reserved.
///////////////////////////////////////////////////////////////////////////////
#pragma once
#include <functoid.h> // BizTalk header for IFunctoid
#include <vector>

#ifndef _CPPUNWIND
#pragma message("***************************************************************************************************")
#pragma message("*** Exception handling must be enabled for the STL classes used here.")
#pragma message("***************************************************************************************************")
#endif


// IFunctoidDispatchImpl<T,L>
// --------------------------
//  provides the IDispatch implementation of the BizTalk interface IFunctoid.
//
//  This class inherits from IDipatchImpl<T, &__uuidof(T), L).
//
//  The implementation is driven from meta data defined by macros BEGIN_FUNCTOID_MAP/END_FUNCTOID_MAP.
//
//  class T uses this class to implement IFunctoid as follows.
//      1) inherit from IFunctoidDispatchImpl<T,L>
//          where L = type library guid for the coclass defined by T
//      2) uses the macro DECLARE_FUNCTOID_MAP()
//      3) uses the macros BEGIN_FUNCTOID_MAP/END_FUNCTOID_MAP in the .cpp file
//      4) expose IFunctoid in the COM_MAP
//      5) expose IDispatch in the COM_MAP to be the IFunctoid implementation
//      6) implement a set of callback functions to be called by IFunctoid::GetScriptBuffer().
//         The names of the callback functions appear in FUNCTOID_IMPL macros nested
//         within the BEGIN_FUNCTOID_MAP/END_FUNCTOID_MAP macros.
//
//  Example:
//      class ATL_NO_VTABLE CAbc : 
//      	public CComObjectRootEx<CComSingleThreadModel>,
//          public CComCoClass<CAbc, &CLSID_Abc>,
//          public IDispatchImpl<IAbc, &IID_IAbc, &LIBID_ABCLib>,
//          public IFunctoidDispatchImpl<CAbc, &LIBID_ABCLib>
//          {
//          public:
//          	CAbc()
//          	{
//          	}
//          
//          DECLARE_REGISTRY_RESOURCEID(IDR_ABC)
//          
//          DECLARE_PROTECT_FINAL_CONSTRUCT()
//          
//          BEGIN_COM_MAP(CAbc)
//          	COM_INTERFACE_ENTRY(IFunctoid)
//          	COM_INTERFACE_ENTRY2(IDispatch, IFunctoid)
//          	COM_INTERFACE_ENTRY(IAbc)
//          END_COM_MAP()
//          
//          public:
//              DECLARE_FUNCTOID_MAP()
//              // IFunctoid::GetScriptBuffer() callbacks.  One for each occurrence
//              // of the FUNCTOID_IMPL macro in the functoid map in the .cpp file.
//              BSTR GetFunctoidImpl1 (long nParameters); 
//              BSTR GetFunctoidImpl2 (long nParameters);          
//          };
//
//      in the .cpp file:
//          BEGIN_FUNCTOID_MAP(CAbc, 2000, 1)
//                  etc.
//          END_FUNCTOID_MAP
//
template <class T, const GUID* plibid = &CComModule::m_libid>
class ATL_NO_VTABLE IFunctoidDispatchImpl
:   public IDispatchImpl<IFunctoid, &__uuidof(IFunctoid), plibid>
{
// construction/destruction
public:
    IFunctoidDispatchImpl ( void )
    :   m_nFunctoidVersion(1), m_nMinFuncId(1001/*0-1000 reserved by BizTalk*/), m_functoids()
    {   // invoke the initalizer declared in class T using
        //  BEGIN_FUNCTOID_MAP/END_FUNCTOID_MAP
        ((T*)this)->buildFunctoidMap();
    }
    virtual ~IFunctoidDispatchImpl ( void )
    {   // delete all dynamically allocated functoids
        for ( unsigned i = 0;  i < m_functoids.size();  ++ i )
            delete m_functoids[i];
    }

// map-based implementation
protected:
    typedef T _FunctoidOuter; // _FunctoidOuter is needed by macro FUNCTOID_IMPl
    struct Functoid // simple struct for all metadata maintained for each functoid
    {   const wchar_t * m_pwszName;          // function name to appear in BizTalk Mapper
        const wchar_t * m_pwszTooltip;       // function tooltip to appear in BizTalk Mapper
        FUNC_CATEGORY m_nCategory;           // function category; defines which tab on the functoid palette to use in the BizTalk Mapper
        CONNECTION_TYPE m_out;               // connection type for the required output parameter
        CONNECTION_TYPE m_inoptional;        // connection type to use for all optional input parameters; CONNECT_TYPE_NONE means fixed number of input parameters
        std::vector<CONNECTION_TYPE> m_in;   // list of connection types for input parameters, one per parameter
        SCRIPT_CATEGORY m_nScriptCategory;   // script category (vbscript or jscript)
        DWORD m_dwBitmapId;                  // resource id of the 16x16 bitmap to appear in BizTalk Mapper
        BSTR (T::* m_pfnGetScript) ( long nParameters ); // callback to class T to calculate script
        Functoid ( const wchar_t * pwszName, FUNC_CATEGORY nCategory )
            :   m_pwszName(pwszName), m_nCategory(nCategory), m_pwszTooltip(L""),
                m_nScriptCategory(SCRIPT_CATEGORY_VBSCRIPT), m_dwBitmapId(0), m_pfnGetScript(0),
                m_out(CONNECT_TYPE_ALL), m_in(), m_inoptional(CONNECT_TYPE_NONE /*flags no optional parameters*/)
            {
            }
    };

    unsigned m_nFunctoidVersion;         // version of this functoid group
    FUNCID m_nMinFuncId;                 // minimum function id, F, used in this group; ids F, F+1, F+2, ..., F+N-1 will be used, N=m_functoids.size()
    std::vector<Functoid*> m_functoids;  // list of functoid descriptors, one per functoid

// IFunctoid implementation
//      Implementation of each interface member uses the metadata above.  This metadata
//      is initialized by the BEGIN_FUNCTOID_MAP/END_FUNCTOID_MAP macros.
public:
    // IFunctoid::get_Version() -- recorded in member m_nFunctoidVersion which is
    //      intialized in macro BEGIN_FUNCTOID_MAP
    STDMETHODIMP get_Version(long * pnVersion)
    {
        if (! pnVersion) return E_POINTER;
        *pnVersion = m_nFunctoidVersion;
        return S_OK;
    }

    // IFunctoid::get_FunctionsCount() -- returns the length of member array m_functoids.
    //      Initalized implicitly through the use of BEGIN_FUNCTOID macros.  Each macro
    //      appends one entry to the m_functoids array.
    STDMETHODIMP get_FunctionsCount(long *pnCount)
    {
        if (! pnCount) return E_POINTER;
        *pnCount = (long) m_functoids.size();
        return S_OK;
    }

    // IFunctoid::GetFunctionsDescripter() -- returns all the metadata for the specific
    //      functoid.  The given index is in the range 0..N-1 where N = number of functoids
    //      returned by get_FunctionsCount().
    //
    //      Most output parameters are recorded directly in m_functoids[i], i = first argument.
    //      pnFuncType = FUNC_TYPE_STD unless there are a variable number of input parameters
    //                   in which case FUNC_TYPE_VARIABLE_INPUT is used.  There are a variable
    //                   number of inputs iff functoids[i].m_inoptional is not CONNECT_TYPE_NONE.
    //      pnFuncId   = m_nMinFuncId + i
    STDMETHODIMP GetFunctionDescripter(long nIndex, FUNC_CATEGORY * pnFuncCategory, SCRIPT_CATEGORY * pnScriptCategory, FUNC_TYPE * pnFuncType, BSTR * pbstrName, BSTR * pbstrTooltip, long * pnBitmapId, long * pnParamCount, FUNCID * pnFuncId)
    {
        if (!pnFuncCategory || !pnScriptCategory || !pnFuncType || !pbstrName || !pbstrTooltip || !pnBitmapId || !pnParamCount || !pnFuncId) return E_POINTER;
        if ((nIndex < 0) || (nIndex >= (long)m_functoids.size())) return E_INVALIDARG;
        const Functoid & f = * m_functoids[nIndex];
        *pnFuncCategory   = f.m_nCategory;
        *pnScriptCategory = f.m_nScriptCategory;
        *pnFuncType       = (f.m_inoptional == CONNECT_TYPE_NONE) ? FUNC_TYPE_STD : FUNC_TYPE_VARIABLEINPUT;
        *pbstrName        = ::SysAllocString(f.m_pwszName);
        *pbstrTooltip     = ::SysAllocString(f.m_pwszTooltip);
        *pnBitmapId       = (long) f.m_dwBitmapId;
        *pnParamCount     = (long) f.m_in.size();
        *pnFuncId         = (FUNCID) (nIndex + m_nMinFuncId);
        return S_OK;
    }

    // IFunctoid::GetFunctionsParameter() -- returns the connection type used for the given
    //      parameter of the given functoid.
    //      The FUNCID given is converted to an integer index, i, in the range 0..N-1 where
    //      N is the number of functoids.  This conversion is done by subtracting m_nMinFuncId
    //      to reverse the algorithm used by GetFunctionDescripter() in construction ids.
    //      
    //      Let p = index of the parameter.
    //      if p is -1, then it refers to the functoid's output parmaeter.  The value returned
    //      for this case is m_functoids[i].m_out.  Macro FUNCTOID_OUT defines this value.
    //      If p index a fixed input parameters, the value returned is in m_functoids[i].m_in[j].
    //      A series of FUNCTOID_IN macros define this array.
    //      If p exceeds the number of fixed parameters and the functoid supports variable
    //      length input parameters, then the value returned is in m_functoids[i].m_inoptional.
    //      Macor FUNCTOID_IN_OPTIONAL defines this value.
    STDMETHODIMP GetFunctionParameter(FUNCID nFuncId, long nParameter, long * pnConnectionType)
    {
        if (! pnConnectionType) return E_POINTER;
        FUNCID nMaxFuncId = (FUNCID) ((long)m_nMinFuncId + (long)m_functoids.size() - 1);
        if ((nFuncId < m_nMinFuncId) || (nFuncId > nMaxFuncId)) return E_INVALIDARG;
        long nIndex = (long)nFuncId - (long)m_nMinFuncId;
        const Functoid & f = * m_functoids[nIndex];
        if (nParameter == -1) // output parameter
        {
            *pnConnectionType = (long) f.m_out;
            return S_OK;
        }
        if ((nParameter < 0) || (nParameter >= (long) f.m_in.size())) return E_INVALIDARG;
        *pnConnectionType = f.m_in[nParameter];
        return S_OK;
    }

    // IFunctoid::GetScriptBuffer() -- returns the script used for the given functoid.  The
    //      number of parameters to be passed to the functoid is given so that the calculation
    //      of the script can account for the parameter list directly.
    //
    //      This implementation delegates to a member function of T.  There is a pointer to
    //      the member function in m_functoids[i].m_pfnGetScript.  The index i is computed
    //      from the FUNCID using the technique used in GetFunctionParameter().
    //
    //      m_pfnGetScript is initialized using the FUNCTOID_IMPL macro
    //
    STDMETHODIMP GetScriptBuffer(FUNCID nFuncId, long nInputParameters, BSTR * pbstrScriptBuffer)
    {
        if (! pbstrScriptBuffer) return E_POINTER;
        FUNCID nMaxFuncId = (FUNCID) ((long)m_nMinFuncId + (long)m_functoids.size() - 1);
        if ((nFuncId < m_nMinFuncId) || (nFuncId > nMaxFuncId)) return E_INVALIDARG;
        long nIndex = (long)nFuncId - (long)m_nMinFuncId;
        const Functoid & f = * m_functoids[nIndex];
        if (nInputParameters < 0) nInputParameters = 0;
        if (f.m_inoptional == CONNECT_TYPE_NONE) // don't have optional number of parameters
            if (nInputParameters > (long) f.m_in.size()) 
                nInputParameters = (long) f.m_in.size();
        BSTR (T::* pfnGetScript)(long) = f.m_pfnGetScript;
        *pbstrScriptBuffer = (((T*)this)->*pfnGetScript)(nInputParameters);
        return S_OK;
    }

};


// FUNCTOID_MAP macros
// --------------------
//  Example usage:
//      BEGIN_FUNCTOID_MAP(CAbc, _VERSION, _FUNCTIONIDBASE)
//         BEGIN_FUNCTOID(L"AbcFunctoid #1", FUNC_CATEGORY_UNKNOWN)
//              FUNCTOID_TOOLTIP(L"Custom functoid with exactly 3 input params.")
//              FUNCTOID_BITMAP(IDB_FUNCTOID_ABCFUNCTOID)
//              FUNCTOID_OUT(CONNECT_TYPE_ALL)
//              FUNCTOID_IN(CONNECT_TYPE_ALL_EXCEPT_RECORD)
//              FUNCTOID_IN(CONNECT_TYPE_ALL_EXCEPT_RECORD) 
//              FUNCTOID_IN(CONNECT_TYPE_ALL_EXCEPT_RECORD)
//              FUNCTOID_IMPL(SCRIPT_CATEGORY_JSCRIPT, GetFunctoidImpl)
//         END_FUNCTOID
//
//        /* other BEGIN_FUNCTOID/END_FUNCTOID entries can follow */
//      END_FUNCTOID_MAP
//
//      These macros expand to implement the member function CAbc::buildFunctoidMap()
//      which is declared in the .h file using DECLARE_FUNCTOID_MAP().

// BEGIN_FUNCTOID_MAP(FI, V, B) -- start the map.  FI=the class inheriting from
//      IFunctoidDispatchImpl.  V=version number and B=minimum functoid id to use.
//      V is recorded in m_nFunctoidVersion and B is recorded in m_nMinFuncId/
//
//      This macro will start implementing function FI::buildFunctoidMap().  Subsequent
//      macros will add to the implemenation.  END_FUNCTOID_MAP will end the implementation
//      by providing the closing curly brace '}'.
#define BEGIN_FUNCTOID_MAP(FI, V, B) \
void FI::buildFunctoidMap ( void ) \
{\
    m_nFunctoidVersion = (unsigned) (V); \
    m_nMinFuncId  = (unsigned) (B);

// END_FUNCTOID_MAP -- provide the closing curly brace to end the implementation of
//  buildFunctoidMap() started with BEGIN_FUNCTOID_MAP.
#define END_FUNCTOID_MAP }

// BEGIN_FUNCTOID(N,C) -- add another functoid functioin to the functoid map.
//      N = name to appear in the BizTalk mapper tool
//      C = function category used to select the tab on the functoid palette in 
//          the BizTalk mapper tool
// This must appear between a BEGIN_FUNCTOID_MAP/END_FUNCTOID_MAP pair.  This
// macro must be closed with an END_FUNCTOID macro.
#define BEGIN_FUNCTOID(N, C) m_functoids.push_back(new Functoid(N, (FUNC_CATEGORY)(C)));

// FUNCTOID_BITMAP(B) -- set the resource id for the functoid's bitmap that will be
//      used in the BizTalk mapper tool.  Used between a BEGIN_FUNCTOID/END_FUNCTOID pair.
#define FUNCTOID_BITMAP(B) m_functoids.back()->m_dwBitmapId = (DWORD)(B);

// FUNCTOID_TOOLTIP(T) -- set the functoid's tooltip in the BizTalk mapper tool.
//      Used between a BEGIN_FUNCTOID/END_FUNCTOID pair.
#define FUNCTOID_TOOLTIP(T) m_functoids.back()->m_pwszTooltip = (const wchar_t *)(T);

// FUNCTOID_OUT(C) -- set connection type for the functoid's single output parameter.
//      Used between a BEGIN_FUNCTOID/END_FUNCTOID pair.
#define FUNCTOID_OUT(C) m_functoids.back()->m_out = (CONNECTION_TYPE)(C);

// FUNCTOID_IN(C) -- set connection type for one of the functoid's input parameters.
//      One macro is used for each fixed input parameter.
//      Used between a BEGIN_FUNCTOID/END_FUNCTOID pair.
#define FUNCTOID_IN(C) m_functoids.back()->m_in.push_back((CONNECTION_TYPE)(C));

// FUNCTOID_IN_OPTIONAL(C) -- set connection type for all of the functoid's optional
//      input parameters, if any.  Used between a BEGIN_FUNCTOID/END_FUNCTOID pair.
#define FUNCTOID_IN_OPTIONAL(C) m_functoids.back()->m_inoptional = (CONNECTION_TYPE)(C);

// FUNCTOID_IMPL(S,F) -- specifies the implementation for the functoid.  S = the script
//      type and must be jscript or vbscript.  F is member function in the class implementing
//      IFunctoid.  This member function will be called to calculate the script.
//      Used between a BEGIN_FUNCTOID/END_FUNCTOID pair.
#define FUNCTOID_IMPL(S, F) \
    m_functoids.back()->m_nScriptCategory = (SCRIPT_CATEGORY)(S); \
    m_functoids.back()->m_pfnGetScript = & _FunctoidOuter::F;

// END_FUNTOID -- ends the speicfication of a functoid.  Used to close a BEGIN_FUNCTOID macro.
//      After this macro, another BEGIN_FUNCTOID macro can appear or the closing
//      END_FUNCTOID_MAP macro may appear.
#define END_FUNCTOID

// DECLARE_FUNCTOID_MAP() -- used in the .h file in the class that inherits from 
//      IFunctoidDispatchImpl.  This macro declares the function buildFunctoidMap()
//      which is implemented by the BEGIN_FUNCTIOD_MAP/END_FUNCTOID_MAP family of macros.
#define DECLARE_FUNCTOID_MAP() void buildFunctoidMap ( void );

// FUNCTOID_FIRST_USER_FUNCID -- is the first available FUNCID for user-defined functoids.
//      The BizTalk SDK reserves FUNCIDs 0 - 1000 for BizTalk.  The first availiable
//      value is 1001.  This macro can be used as a parameter to the BEGIN_FUNCTOID_MAP macro.
#define FUNCTOID_FIRST_USER_FUNCID 1001
